import pytest
from astropy.cosmology import Planck15

from horizonground.lumfunc_modeller import (
    LumFuncModeller,
    alpha_emitter_schechter_lumfunc,
    quasar_hybrid_lumfunc,
    quasar_hybrid_model_constraint,
    quasar_PLE_lumfunc,
    quasar_PLE_model_constraint,
)


@pytest.mark.parametrize(
    "m,z,lg_phi",
    [
        (-27., 1., -7.99),
        (-26., 2., -5.92),
    ]
)
def test_quasar_PLE_lumfunc(quasar_PLE_model_params, m, z, lg_phi):

    assert quasar_PLE_lumfunc(m, z, **quasar_PLE_model_params) \
        == pytest.approx(lg_phi, abs=0.01)


@pytest.mark.parametrize(
    "m,z,lg_phi",
    [
        (-27., 1., -7.88),
        (-26., 2., -5.92),
        (-25., 3.25, -6.22),
    ]
)
def test_quasar_hybrid_lumfunc(quasar_PLE_LEDE_model_params, m, z, lg_phi):

    assert quasar_hybrid_lumfunc(m, z, **quasar_PLE_LEDE_model_params) \
        == pytest.approx(lg_phi, abs=0.01)


def test_quasar_PLE_model_constraint(quasar_PLE_model_params):

    assert quasar_PLE_model_constraint(**quasar_PLE_model_params)

    # Exchange power-law indices to break the constraint.
    quasar_PLE_model_params.update({
        r'\alpha_\mathrm{l}': quasar_PLE_model_params[r'\beta_\mathrm{l}'],
        r'\beta_\mathrm{l}': quasar_PLE_model_params[r'\alpha_\mathrm{l}'],
    })
    assert not quasar_PLE_model_constraint(**quasar_PLE_model_params)


def test_quasar_hybrid_model_constraint(quasar_PLE_LEDE_model_params):

    assert quasar_hybrid_model_constraint(**quasar_PLE_LEDE_model_params)

    # Exchange power-law indices to break the constraint.
    quasar_PLE_LEDE_model_params.update({
        r'\alpha': quasar_PLE_LEDE_model_params[r'\beta'],
        r'\beta': quasar_PLE_LEDE_model_params[r'\alpha'],
    })
    assert not quasar_hybrid_model_constraint(**quasar_PLE_LEDE_model_params)


class TestLumFuncModeller:

    @pytest.mark.parametrize(
        "z,lumfunc,lumparams,lumvar,threshold,cosmology,value",
        [
            (
                2.0,
                quasar_PLE_lumfunc,
                'quasar_PLE_model_params',
                'magnitude',
                -22.,
                Planck15,
                5.91e-5,
            ),
            (
                2.0,
                alpha_emitter_schechter_lumfunc,
                'alpha_emitter_schechter_model_params',
                'luminosity',
                3.e-16,
                Planck15,
                5.63e-5,
            ),
        ]
    )
    def test_comoving_number_density(self, quasar_PLE_model_params,
                                     alpha_emitter_schechter_model_params,
                                     z, lumfunc, lumparams, lumvar, threshold,
                                     cosmology, value):

        lumparams = locals()[lumparams]

        test_instance = LumFuncModeller(
            lumfunc, lumparams, lumvar, threshold, cosmology=cosmology,
            normalisation_redshift=2.
        )

        assert test_instance.comoving_number_density(z) \
            == pytest.approx(value, rel=0.01)


    @pytest.mark.parametrize(
        "z,lumfunc,lumparams,lumvar,threshold,cosmology,value",
        [
            (
                2.0,
                quasar_PLE_lumfunc,
                'quasar_PLE_model_params',
                'magnitude',
                -22.,
                Planck15,
                -0.332,
            ),
            (
                2.0,
                alpha_emitter_schechter_lumfunc,
                'alpha_emitter_schechter_model_params',
                'luminosity',
                3.e-16,
                Planck15,
                -7.65,
            ),
        ]
    )
    def test_evolution_bias(self, quasar_PLE_model_params,
                            alpha_emitter_schechter_model_params,
                            z, lumfunc, lumparams, lumvar, threshold,
                            cosmology, value):

        lumparams = locals()[lumparams]

        test_instance = LumFuncModeller(
            lumfunc, lumparams, lumvar, threshold, cosmology=cosmology,
            normalisation_redshift=2.
        )

        assert test_instance.evolution_bias(z) \
            == pytest.approx(value, rel=0.01)


    @pytest.mark.parametrize(
        "z,lumfunc,lumparams,lumvar,threshold,cosmology,value",
        [
            (
                2.0,
                quasar_PLE_lumfunc,
                'quasar_PLE_model_params',
                'magnitude',
                -22.,
                Planck15,
                0.0748,
            ),
            (
                2.0,
                alpha_emitter_schechter_lumfunc,
                'alpha_emitter_schechter_model_params',
                'luminosity',
                3.e-16,
                Planck15,
                0.54,
            ),
        ]
    )
    def test_magnification_bias(self, quasar_PLE_model_params,
                                alpha_emitter_schechter_model_params,
                                z, lumfunc, lumparams, lumvar, threshold,
                                cosmology, value):

        lumparams = locals()[lumparams]

        test_instance = LumFuncModeller(
            lumfunc, lumparams, lumvar, threshold, cosmology=cosmology,
            normalisation_redshift=2.
        )

        assert test_instance.magnification_bias(z) \
            == pytest.approx(value, rel=0.01)



"""
Utilities (:mod:`~horizonground.utils`)
===========================================================================

Utilities tools.

.. autosummary::

    process_header
    load_parameter_set

|

"""
import os
from pathlib import Path

__all__ = [
    'process_header',
    'load_parameter_set',
    'get_test_data_loc',
]


def process_header(header, skipcols=0):
    r"""Process comment-line header (usually indicated by the \'#\'
    character) of text files.

    Parameters
    ----------
    header : str
        File header line either delimited by comma or space.
    skipcols : int, optional
        Skip the first columns (default is 0).

    Returns
    -------
    columns : list of str
        Column headings of the file.

    """
    header = header.strip("#").strip("\n")

    if "," in header:
        columns = list(map(
            lambda heading: heading.strip(), header.split(",")[skipcols:]
        ))
    else:
        columns = [
            heading.strip()
            for heading in header.split()[skipcols:] if not heading.isspace()
        ]

    return columns


def load_parameter_set(parameter_file):
    r"""Load a parameter set from a file into a dictionary.

    Parameters
    ----------
    parameter_file : *str or* :class:`pathlib.Path`
        Parameter file with the header line giving the parameter names
        and the first non-comment line giving the parameter values
        delimited by comma.

    Returns
    -------
    parameter_set : dict
        Parameter set.

    Notes
    -----
    Parameters with names beginning with '\Delta' are discarded.

    """
    with open(parameter_file, 'r') as pfile:
        parameters = process_header(pfile.readline())
        estimates = tuple(map(float, pfile.readline().split(",")))

    parameter_set = dict(zip(parameters, estimates))
    for parameter in parameters:
        if parameter.startswith(r"\Delta"):
            del parameter_set[parameter]

    return parameter_set


def get_test_data_loc(filename):
    """Get location of the test data set.

    Parameters
    ----------
    filename : str
        File name of the test data set including its extension.

    Returns
    -------
    filepath : :class:`pathlib.Path`
        Path to the test data file.

    """
    pkg_dir = os.path.dirname(os.path.abspath(__file__))

    filepath = Path(pkg_dir)/"tests"/"test_data"/filename

    return filepath



"""Test configuration for :mod:`horizonground`.

"""
import os
import sys
from pathlib import Path

import pytest

from .utils import load_parameter_set

current_file_dir = Path(os.path.dirname(os.path.abspath(__file__)))
test_data_dir = param_file_path = current_file_dir/"tests"/"test_data"


def pytest_addoption(parser):
    """Add command-line options to `pytest` parser.

    Parameters
    ----------
    parser : :class:`_pytest.config.argparsing.Parser`
        `pytest` parser object.

    """
    parser.addoption(
        '--runslow', action='store_true', default=False, help="run slow tests"
    )


def pytest_configure(config):
    """Add ini-file options to `pytest` configuration.

    Parameters
    ----------
    config : :class:`_pytest.config.Config`
        `pytest` configuration object.

    """
    config.addinivalue_line('markers', "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    """Modify test collection items.

    Parameters
    ----------
    config : :class:`_pytest.config.Config`
        `pytest` configuration object.
    items : list of :class:`_pytest.nodes.Item`
        `pytest` item objects.

    """
    if config.getoption('--runslow'):
        return

    skip_slow = pytest.mark.skip(
        reason="use --runslow option to run slow tests"
    )
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture(scope="session", autouse=True)
def quasar_PLE_model_params(request):

    return load_parameter_set(
        test_data_dir/"eBOSS_QSO_LF_PLE_model_fits.txt"
    )


@pytest.fixture(scope="session", autouse=True)
def quasar_PLE_LEDE_model_params(request):

    return load_parameter_set(
        test_data_dir/"eBOSS_QSO_LF_PLE+LEDE_model_fits.txt"
    )


@pytest.fixture(scope="session", autouse=True)
def alpha_emitter_schechter_model_params(request):

    return load_parameter_set(
        test_data_dir/"H-alpha_LF_Schechter_model_fits.txt"
    )



r"""
Luminosity function modeller (:mod:`~horizonground.lumfunc_modeller`)
===========================================================================

Provide models of the redshift-dependent tracer luminosity function
:math:`\Phi(m, z)` (for absolute magnitude :math:`m`) or
:math:`\Phi(\lg{L}, z)` (for base-10 logarithm of the intrinsic
luminosity :math:`L`), from which the comoving number density below/above
some brightness threshold :math:`\bar{m}` or :math:`\bar{L}`,

.. math::

    \bar{n}(z; <\!\bar{m}) = \int_{-\infty}^{\bar{m}}
        \operatorname{d}\!m\, \Phi(m, z)
    \quad \mathrm{or} \quad
    \bar{n}(z; >\!\lg\bar{L}) = \int^{\infty}_{\lg\bar{L}}
        \operatorname{d}\!\lg{L}\, \Phi(\lg{L}, z) \,,

can be predicted.  The corresponding evolution bias is

.. math::

    b_\mathrm{e}(z) = - (1 + z)
        \frac{\partial \ln\bar{n}(z)}{\partial z}

and magnification bias is

.. math::

    s(z) = \frac{1}{\ln10}
        \frac{\Phi(\bar{m},z)}{\bar{n}(z; <\!\bar{m})}
    \quad \mathrm{or} \quad
    s(z) = \frac{2}{5\ln10}
        \frac{\Phi(\lg\bar{L},z)}{\bar{n}(z; >\!\lg\bar{L})}
    \,,

We also offer a simple :math:`K`-correction formula taken from ref. [1]_.

.. [1] Croom S. M. et al., 2009. **MNRAS** 399(4), 1755--1772.
   [arXiv: `0907.2727 <https://arxiv.org/abs/0907.2727>`_]


.. autosummary::

    LumFuncModeller
    konstante_correction


Quasar (QSO) luminosity function
---------------------------------------------------------------------------

**Pure luminosity evolution model (PLE)**

The quasar luminosity function in the pure luminosity evolution (PLE)
model is a double power law

.. math::

    \Phi(m, z) = \frac{\Phi_\ast}{
        10^{0.4 (\alpha + 1) [m - m_\ast(z)]}
        + 10^{0.4 (\beta + 1) [m - m_\ast(z)]}
    }

where :math:`m` is the absolute magnitude, :math:`z` is the redshift, and
the slope parameters :math:`\alpha, \beta` describe the power laws on the
bright and faint ends respectively, but their values differ below and
above the pivot redshift :math:`z_\mathrm{p}`.  :math:`m_\ast` is the
break magnitude at which the luminosity function evaluates to
:math:`\Phi_\ast`,

.. math::

    m_\ast(z) = m_\ast(z_\mathrm{p}) - \frac{5}{2} \left[
        k_1 (z - z_\mathrm{p}) + k_2 (z - z_\mathrm{p})^2
    \right] \,,

where :math:`k_1, k_2` are the redshift-evolution parameters whose values
also differ below and above the pivot redshift [2]_.

This is a parametric model with 10 parameters: :math:`\lg\Phi_\ast`,
:math:`m_\ast(z_\mathrm{p})`, :math:`(\alpha, \beta, k_1, k_2)_\mathrm{l}`
for :math:`z < z_\mathrm{p}` and :math:`(\alpha, \beta, k_1,
k_2)_\mathrm{h}` for :math:`z > z_\mathrm{p}`.   Due to the exchange
symmetry between :math:`\alpha` and :math:`\beta` in the double power law,
the model constraint :math:`\alpha < \beta` is imposed for both above
and below the pivot redshift.


**Hybrid evolution model (PLE+LEDE)**

This model is the same as the PLE model below the pivot redshift, but
above that the luminosity evolution--density evolution (LEDE) model is
adopted where the luminosity function normalisation, break magnitude and
bright-end power law index have different redshift evolutions

.. math::

    \begin{align*}
        \lg\Phi_\ast &= \lg\Phi_\ast(z_\mathrm{p}) \
            + c_{1\mathrm{a}} (z - z_\mathrm{p})
            + c_{1\mathrm{b}} (z - z_\mathrm{p})^2 \,, \\
        m_\ast(z) &= m_\ast(z_\mathrm{p}) + c_2 (z - z_\mathrm{p}) \,, \\
        \alpha(z) &= \alpha(z_\mathrm{p}) + c_3 (z - z_\mathrm{p}) \,,
    \end{align*}

and continuity across redshift is imposed by requiring the same
:math:`\lg\Phi_\ast(z_\mathrm{p})` and :math:`m_\ast(z_\mathrm{p})` for
two models [2]_.

The hybrid model still has 10 overall parameters: the PLE model retains 6
low-redshift parameters and the high-redshift LEDE model has 8 parameters,
with the substitutions of :math:`\lg\Phi_\ast(0)` for :math:`\lg\Phi_\ast`,
:math:`m_\ast(0)` for :math:`m_\ast(z_\mathrm{p})`, :math:`c_2` for
:math:`k_1, k_2` and the addition of
:math:`c_{1\mathrm{a}}, c_{1\mathrm{b}}` and :math:`c_3`.  As before, due
to the exchange symmetry between :math:`\alpha` and :math:`\beta`, the
low-redshift PLE model constraint :math:`\alpha < \beta` is imposed.

.. [2] Palanque-Delabrouille N. et al., 2016. **A&A** 587, A41.
   [arXiv: `1509.05607 <https://arxiv.org/abs/1509.05607>`_]


.. autosummary::

    quasar_PLE_lumfunc
    quasar_PLE_model_constraint
    quasar_hybrid_lumfunc
    quasar_hybrid_model_constraint


H |alpha| -emitter luminosity function
---------------------------------------------------------------------------

**Schechter function model**

The H |alpha| -emitter luminosity function in the Schechter function model
takes the form of a gamma function

.. math::

    \Phi(L, z) \operatorname{d}\!L = \underbrace{
        \ln10 \, \Phi_\ast(z) y(z)^{\alpha+1} \mathrm{e}^{-y(z)}
    }_{\Phi(\lg{L}, z)} \operatorname{d}\!\lg{L}

where :math:`\alpha` is the faint-end slope parameter,

.. math::

    \begin{align*}
        \Phi_\ast(z) &=
            \begin{cases}
                \Phi_{\ast0} (1 + z)^\epsilon \,,
                    \quad z \leqslant z_\mathrm{b} \,; \\
                \Phi_{\ast0} (1 + z_\mathrm{b})^{2\epsilon}
                    (1 + z)^{-\epsilon} \,, \quad z > z_\mathrm{b} \,,
            \end{cases} \\
        y(z) &= \frac{L}{L_{\ast0}} (1 + z)^{-\delta} \,,
    \end{align*}

are the redshift-dependent characteristic comoving number density and
relative luminosity of the H |alpha| -emitters, :math:`\epsilon, \delta`
are the redshift-evolution indices, and :math:`z_\mathrm{b}` is the break
magnitude [3]_.

This is a parametric model with 6 parameters: :math:`\alpha, \epsilon,
\delta`, :math:`z_\mathrm{b}`, :math:`m_{\ast0}` and :math:`\Phi_{\ast0}`.

.. [3] Pozzetti L. et al., 2016. **A&A** 590, A3.
   [arXiv: `1603.01453 <https://arxiv.org/abs/1603.01453>`_]


.. autosummary::

    alpha_emitter_schechter_lumfunc

|

.. |alpha| unicode:: U+03B1
    :trim:

"""
from inspect import signature

import numpy as np
from astropy import units
from scipy.integrate import quad
from scipy.misc import derivative


def konstante_correction(redshift, normalisation_redshift=2., index=-0.5):
    r"""Subtractive Konstante correction (*K*-correction) to apparent
    magnitude for correcting bandpass redshifting.

    Notes
    -----
    This implements the subtractive *K*-correction :math:`K(z) - K(z_\ast)`
    normalised to redshift :math:`z_\ast`, where

    .. math::

        K(z) = - \frac{5}{2} (1 + \alpha_\nu) \lg(1 + z)

    and :math:`\alpha_\nu` is the power-law index.

    Parameters
    ----------
    redshift : float
        Redshift.
    normalisation_redshift : float, optional
        Normalisation redshift (default is 2.).
    index : float, optional
        :math:`K`-correction power-law index (default is -0.5).

    Returns
    -------
    correction_magnitude : float
        The magnitude correction that needs to be subtracted from the
        observed magnitude.

    """
    return - 5./2. * (1 + index) \
        * np.log10((1 + redshift) / (1 + normalisation_redshift))


def quasar_PLE_lumfunc(magnitude, redshift, *, base10_log=True,
                       redshift_pivot=2.2, **model_parameters):
    r"""Evaluate the pure luminosity evolution (PLE) model for the quasar
    luminosity function at the given absolute magnitude and redshift.

    Parameters
    ----------
    magnitude : float
        Quasar absolute magnitude.
    redshift : float
        Quasar redshift.
    base10_log : bool, optional
        If `True` (default), return the base-10 logarithmic value.
    redshift_pivot : float, optional
        Pivot redshift.
    **model_parameters
        PLE model parameters.  Must be passed with the following
        parameter names:
        ``r'\lg\Phi_\ast'``, ``r'm_\ast(z_\mathrm{p})'``,
        ``r'\alpha_\mathrm{{l}}'``, ``r'\beta_\mathrm{{l}}'``,
        ``r'k_{{1\mathrm{{l}}}}'``, ``r'k_{{2\mathrm{{l}}}}'``,
        ``r'\alpha_\mathrm{{h}}'``, ``r'\beta_\mathrm{{h}}'``,
        ``r'k_{{1\mathrm{{h}}}}'``, ``r'k_{{2\mathrm{{h}}}}'``.

    Returns
    -------
    lumfunc_value : float
        Predicted qausar luminosity function value (in inverse cubic Mpc
        per unit magnitude).  Base-10 logarithmic value is returned if
        `base10_log` is `True`.

    Notes
    -----
    This function is not vectorised.  Please use `numpy.vectorize` or
    `functools` if you need vectorisation.

    """
    if not isinstance(model_parameters, dict):
        raise TypeError("Model parameters must be passed as a dictionary.")

    m, z, z_p = magnitude, redshift, redshift_pivot

    # Determine the redshift end for setting parameters.
    subscript = r'\mathrm{{{}}}'.format('l') if z <= z_p \
        else r'\mathrm{{{}}}'.format('h')

    alpha = model_parameters[r'\alpha_{}'.format(subscript)]
    beta = model_parameters[r'\beta_{}'.format(subscript)]
    k_1 = model_parameters[r'k_{{1{}}}'.format(subscript)]
    k_2 = model_parameters[r'k_{{2{}}}'.format(subscript)]

    lg_Phi_star = model_parameters[r'\lg\Phi_\ast']
    m_star_at_z_p = model_parameters[r'm_\ast(z_\mathrm{p})']

    # Evaluate the model prediction.
    magnitude_deviation_exponent = \
        (m - m_star_at_z_p) + 2.5 * (k_1 * (z - z_p) + k_2 * (z - z_p) ** 2)

    ln_faint_power_law = \
        np.log(10) * (0.4 * (alpha + 1) * magnitude_deviation_exponent)
    ln_bright_power_law = \
        np.log(10) * (0.4 * (beta + 1) * magnitude_deviation_exponent)

    ln_denominator = np.logaddexp(ln_faint_power_law, ln_bright_power_law)

    if base10_log:
        lumfunc_value = lg_Phi_star - ln_denominator / np.log(10)
    else:
        Phi_star = 10 ** lg_Phi_star
        lumfunc_value = Phi_star / np.exp(ln_denominator)

    return lumfunc_value


def quasar_PLE_model_constraint(**model_parameters):
    """Check whether the pure luminosity evolution (PLE) model constraint
    is satisfied.

    Parameters
    ----------
    **model_parameters
        PLE model parameters.  See :func:`quasar_PLE_lumfunc` for required
        parameters.

    Returns
    -------
    bool
        Whether or not the model constraint is satisfied.

    """
    return (
        model_parameters[r'\alpha_\mathrm{l}']
        < model_parameters[r'\beta_\mathrm{l}']
    ) and (
        model_parameters[r'\alpha_\mathrm{h}']
        < model_parameters[r'\beta_\mathrm{h}']
    )


def quasar_hybrid_lumfunc(magnitude, redshift, *, base10_log=True,
                          redshift_pivot=2.2, **model_parameters):
    r"""Evaluate the hybrid model (pure luminosity evolution and luminosity
    evolution--density evolution, 'PLE+LEDE') for the quasar luminosity
    function at the given absolute magnitude and redshift.

    Parameters
    ----------
    magnitude : float
        Quasar absolute magnitude.
    redshift : float
        Quasar redshift.
    base10_log : bool, optional
        If `True` (default), return the base-10 logarithmic value.
    redshift_pivot : float, optional
        Pivot redshift.
    **model_parameters
        Hybrid model parameters.  Must be passed with the following
        parameter names: ``r'\lg\Phi_\ast(0)'``, ``r'm_\ast(0)'``,
        ``r'\alpha'``, ``r'\beta'``, ``r'k_1'``, ``r'k_2'``,
        ``r'c_{{1\mathrm{{a}}}}'``, ``r'c_{{1\mathrm{{b}}}}'``,
        ``r'c_2'``, ``r'k_3'``.

    Returns
    -------
    lumfunc_value : float
        Predicted qausar luminosity function value (in inverse cubic Mpc
        per unit magnitude).  Base-10 logarithmic value is returned if
        `base10_log` is `True`.

    Notes
    -----
    This function is not vectorised.  Please use `numpy.vectorize` or
    `functools` if you need vectorisation.

    """
    if not isinstance(model_parameters, dict):
        raise TypeError("Model parameters must be passed as a dictionary.")

    m, z, z_p = magnitude, redshift, redshift_pivot

    # Low redshift PLE model.
    lg_Phi_star_0 = model_parameters[r'\lg\Phi_\ast(0)']
    m_star_0 = model_parameters[r'm_\ast(0)']
    alpha = model_parameters[r'\alpha']
    beta = model_parameters[r'\beta']
    k_1 = model_parameters[r'k_1']
    k_2 = model_parameters[r'k_2']

    if z <= z_p:

        magnitude_deviation_exponent = \
            (m - m_star_0) + 2.5 * (k_1 * z + k_2 * z ** 2)

        ln_faint_power_law = \
            np.log(10) * (0.4 * (alpha + 1) * magnitude_deviation_exponent)
        ln_bright_power_law = \
            np.log(10) * (0.4 * (beta + 1) * magnitude_deviation_exponent)

        ln_denominator = np.logaddexp(ln_faint_power_law, ln_bright_power_law)

        if base10_log:
            lumfunc_value = lg_Phi_star_0 - ln_denominator / np.log(10)
        else:
            Phi_star = 10 ** lg_Phi_star_0
            lumfunc_value = Phi_star / np.exp(ln_denominator)

        return lumfunc_value

    # High redshift LEDE model.
    c_1a = model_parameters[r'c_{1\mathrm{a}}']
    c_1b = model_parameters[r'c_{1\mathrm{b}}']
    c_2 = model_parameters[r'c_2']
    c_3 = model_parameters[r'c_3']

    lg_Phi_star_at_z_p = lg_Phi_star_0
    m_star_at_z_p = m_star_0 - 2.5 * (k_1 * z_p + k_2 * z_p ** 2)

    lg_Phi_star = lg_Phi_star_at_z_p + c_1a * (z - z_p) + c_1b * (z - z_p) ** 2
    magnitude_deviation_exponent = (m - m_star_at_z_p) + c_2 * (z - z_p)
    alpha = alpha + c_3 * (z - z_p)

    ln_faint_power_law = \
        np.log(10) * (0.4 * (alpha + 1) * magnitude_deviation_exponent)
    ln_bright_power_law = \
        np.log(10) * (0.4 * (beta + 1) * magnitude_deviation_exponent)

    ln_denominator = np.logaddexp(ln_faint_power_law, ln_bright_power_law)

    if base10_log:
        lumfunc_value = lg_Phi_star - ln_denominator / np.log(10)
    else:
        Phi_star = 10 ** lg_Phi_star
        lumfunc_value = Phi_star / np.exp(ln_denominator)

    return lumfunc_value


def quasar_hybrid_model_constraint(**model_parameters):
    r"""Check whether the hybrid model constraint is satisfied.

    Parameters
    ----------
    **model_parameters
        Hybrid model parameters.  See :func:`quasar_hybrid_lumfunc` for
        required parameters.

    Returns
    -------
    bool
        Whether or not the model constraint is satisfied.

    """
    return model_parameters[r'\alpha'] < model_parameters[r'\beta']


def alpha_emitter_schechter_lumfunc(luminosity, redshift, *, base10_log=True,
                                    **model_parameters):
    r"""Evaluate the Schechter model for the H |alpha| -emitter
    luminosity function at the given intrinsic luminosity and redshift.

    Parameters
    ----------
    luminosity : float
        H |alpha| -emitter luminosity in dex (base-10 logarithm).
    redshift : float
        H |alpha| -emitter redshift.
    base10_log : bool, optional
        If `True` (default), return the base-10 logarithmic value.
    **model_parameters
        Schechter model parameters.  Must be passed with the following
        parameter names: ``r'\lg\Phi_{\ast0}'``, ``r'\lg{L_{\ast0}}'``,
        ``r'z_\mathrm{b}'``, ``r'\alpha'``, ``r'\delta'``, ``r'\epsilon'``.

    Returns
    -------
    lumfunc_value : float
        Predicted H |alpha| -emitter luminosity function value (in inverse
        cubic Mpc per flux dex).  Base-10 logarithmic value is returned if
        `base10_log` is `True`.

    Notes
    -----
    This function is not vectorised.  Please use `numpy.vectorize` or
    `functools` if you need vectorisation.

    """
    lg_L, z = luminosity, redshift

    lg_Phi_star0 = model_parameters[r'\lg\Phi_{\ast0}']
    lg_L_star0 = model_parameters[r'\lg{L_{\ast0}}']
    z_b = model_parameters[r'z_\mathrm{b}']
    alpha = model_parameters[r'\alpha']
    delta = model_parameters[r'\delta']
    epsilon = model_parameters[r'\epsilon']

    # Evaluate the model prediction.
    if z <= z_b:
        lg_Phi_star = lg_Phi_star0 + epsilon * np.log10(1 + z)
    else:
        lg_Phi_star = lg_Phi_star0 \
            + 2 * epsilon * np.log10(1 + z_b) \
            - epsilon * np.log10(1 + z)

    lg_y = lg_L - lg_L_star0 - delta * np.log10(1 + z)

    if base10_log:
        lumfunc_value = np.log10(np.log(10)) + lg_Phi_star \
            + (alpha + 1) * lg_y - 10 ** lg_y * np.log10(np.e)
    else:
        lumfunc_value = np.log(10) * 10 ** lg_Phi_star \
            * 10 ** (lg_y * (alpha + 1)) * np.exp(- 10 ** lg_y)

    return lumfunc_value


class LumFuncModeller:
    r"""Luminosity function modeller predicting the comoving number
    density, evolution bias and magnification bias for a given brightness
    threshold.

    Parameters
    ----------
    model_lumfunc : callable
        A parametric luminosity function model as a function of brightness
        and redshift (in that order) in units of inverse cubic Mpc.  If it
        returns the luminosity function value in base-10 logarithm,
        `exponentiation` should be set to `True`; if it accepts a boolean
        parameter 'base10_log', this will be overriden to `False`.
    model_parameters : dict
        Model parameters passed to `model_lumfunc` as keyword arguments.
    brightness_variable : {'luminosity', 'magnitude'}, str
        Brightness variable of `model_lumfunc`, either 'luminosity' (in
        dex) or 'magnitude'.
    threshold_value : float
        Brightness threshold value for `brightness_variable`.  If
        `brightness_variable` is 'luminosity', this is interpreted as a
        flux limit and converted to an intrinsic luminosity value at the
        tracer redshift using the luminosity distance.
    normalise_threshold : bool, optional
        If `True` (default is `False`), `threshold_value` is converted
        from a flux limit to a luminosity value at `normalisation_redshift`
        using the luminosity distance if `brightness_variable` is
        'luminosity', or converted from an apparent magnitude limit to
        an absolute magnitude value at `normalisation_redshift` using the
        distance modulus if `brightness_variable` is 'magnitude'.
    normalisation_redshift : float or None, optional
        If not `None` (default) and `normalise_threshold` is `True`, this
        redshift is used convert flux or apparent magnitude limit to an
        intrinsic luminosity or absolute magnitude threshold.
    cosmology : :class:`astropy.cosmology.core.Cosmology` or None, optional
        Background cosmological model used to calculate the luminosity
        distance or distance modulus (default is `None`).
    exponentiation : bool, optional
        If `True` (default is `False`), this assumes `model_lumfunc` only
        returns the base-10 logarithmic luminosity function and
        exponentiate its returned value to a power of 10.

    Attributes
    ----------
    luminosity_function : callable
        Luminosity function of brightness and redshift variables only
        (in that order) (in inverse cubic :math:`\mathrm{Mpc}/h` per unit
        luminosity).
    brightness_threshold : callable
        Intrinsic luminosity or absoloute magnitude threshold, depending
        on the input argument of `brightness_variable`, as a function
        of redshift.
    cosmology : :class:`astropy.cosmology.core.Cosmology`
        Background cosmological model.
    attrs : dict
        Model parameters and luminosity variables stored in a dictionary.

    Notes
    -----
    :attr:`luminosity_function` is a vectorised function.

    """

    luminosity_bound = {
        'luminosity': 100.,
        'magnitude': -50.,
    }
    r"""float: Finite luminosity upper bound for numerical integration.

    If :attr:`brightness_variable` is 'luminosity', the bound is given as
    a base-10 logarithmic flux value in erg/s; else if 'magnitude',
    the bound is dimensionless.

    """

    redshift_stepsize = 0.001
    r"""float: Redshift step size for numerical differentiation.

    """

    def __init__(self, model_lumfunc, model_parameters, brightness_variable,
                 threshold_value, normalise_threshold=False,
                 normalisation_redshift=None, cosmology=None,
                 exponentiation=False):

        self.attrs = {
            'model_parameters': model_parameters,
            'brightness_variable': brightness_variable,
        }

        self.cosmology = cosmology

        if 'base10_log' in signature(model_lumfunc).parameters:
            model_parameters.update({'base10_log': False})
            # pylint: disable=unnecessary-lambda
            self.luminosity_function = np.vectorize(
                lambda lum, z: model_lumfunc(lum, z, **model_parameters)
            )
        elif exponentiation:
            self.luminosity_function = np.vectorize(
                lambda lum, z: 10 ** model_lumfunc(lum, z, **model_parameters)
            )
        else:
            # pylint: disable=unnecessary-lambda
            self.luminosity_function = np.vectorize(
                lambda lum, z: model_lumfunc(lum, z, **model_parameters)
            )

        if brightness_variable == 'luminosity':
            # pylint: disable=no-member
            self.brightness_threshold = lambda z: np.log10(
                4 * np.pi * threshold_value
                * self.cosmology.luminosity_distance(z).to(units.cm).value ** 2
            )
            self._brightness_threshold = \
                self.brightness_threshold(normalisation_redshift)
        elif brightness_variable == 'magnitude':
            self.brightness_threshold = lambda z: \
                threshold_value - self.cosmology.distmod(z).value \
                if normalise_threshold else threshold_value
            self._brightness_threshold = \
                self.brightness_threshold(normalisation_redshift) \
                if normalise_threshold else threshold_value

    @classmethod
    def from_parameter_file(cls, parameter_file, **kwargs):
        """Instantiate the modeller class with model parameter values
        loaded from a file.

        Parameters
        ----------
        parameter_file : *str or* :class:`pathlib.Path`
            Path of the model parameter file.
        **kwargs
            Parameters other than `model_parameters` passed to the
            modeller class as keyword arguments.

        """
        with open(parameter_file, 'r') as pfile:
            parameters = tuple(
                map(
                    lambda var_name: var_name.strip(),
                    pfile.readline().strip("#").strip("\n").split(",")
                )
            )
            estimates = tuple(map(float, pfile.readline().split(",")))

        model_parameters = dict(zip(parameters, estimates))

        return cls(model_parameters=model_parameters, **kwargs)

    def comoving_number_density(self, redshift):
        r"""Return the comoving number density :math:`\bar{n}(z)`
        above the luminosity threshold at a given redshift.

        Parameters
        ----------
        redshift : float
            Redshift.

        Returns
        -------
        float
            Comoving number density (in inverse cubic
            :math:`\mathrm{Mpc}/h`).

        """
        _comoving_number_density = np.abs(
            quad(
                self.luminosity_function,
                self._brightness_threshold,
                self.luminosity_bound[self.attrs['brightness_variable']],
                args=(redshift,)
            )[0]
        ) / self.cosmology.h ** 3

        return _comoving_number_density

    def evolution_bias(self, redshift):
        r"""Return the evolution bias :math:`b_\mathrm{e}(z)` at a
        given redshift.

        Parameters
        ----------
        redshift : float
            Redshift.

        Returns
        -------
        float
            Evolution bias.

        """
        _evolution_bias = - (1 + redshift) * derivative(
            self.comoving_number_density, redshift, dx=self.redshift_stepsize
        ) / self.comoving_number_density(redshift)

        return _evolution_bias

    def magnification_bias(self, redshift):
        r"""Return the magnification bias :math:`s(z)` at a given redshift.

        Parameters
        ----------
        redshift : float
            Redshift.

        Returns
        -------
        float
            Magnification bias.

        """
        if self.attrs['brightness_variable'] == 'luminosity':
            prefactor = 2./5. / np.log(10)
        elif self.attrs['brightness_variable'] == 'magnitude':
            prefactor = 1 / np.log(10)

        _magnification_bias = prefactor * self.luminosity_function(
            self._brightness_threshold, redshift
        ) / self.comoving_number_density(redshift)

        return _magnification_bias



r"""
Luminosity function likelihood (:mod:`~horizonground.lumfunc_likelihood`)
===========================================================================

Fit tracer luminosity functions and produce model constraints.

Data processing
---------------

.. autosummary::

    DataSourceError
    DataSourceWarning
    LumFuncMeasurements


Likelihood evaluation
---------------------

.. autosummary::

    LumFuncLikelihood

|

"""
import warnings
from collections import OrderedDict
from inspect import signature
from itertools import compress as iterfilter
from itertools import product as iterproduct

import numpy as np

from .utils import process_header


class DataSourceError(IOError):
    """Raise an exception when the data source is not available or not in
    the correct format.

    """


class DataSourceWarning(UserWarning):
    """Raise an exception when the data source does not appear to be
    self-consistent.

    """


class LumFuncMeasurements:
    """Luminosity function measurements for a tracer sample.

    Parameters
    ----------
    measurement_file : *str or* :class:`pathlib.Path`
        File path to luminosity function measurements.
    uncertainty_file : *str or* :class:`pathlib.Path` *or None, optional*
        File path to luminosity function uncertainties.
    base10_log : bool, optional
        If `True` (default), measurement values are converted to base-10
        logarithms.  String ``'lg_'`` is detected in file headers to
        determine whether loaded measurements are already in base-10
        logarithms.

    Attributes
    ----------
    luminosity_bins : list of float
        Luminosity bin centres.
    redshift_bins : list of float
        Redshift bin centres.
    redshift_labels : list of str
        Redshift bin labels.

    Notes
    -----
    Data files are assumed to be an array where rows correspond to
    increasing luminosity bins and columns correspond to increasing
    redshift bins.  The first line should be a comment line for column
    headings and the first column should be the luminosity bins.  Redshift
    bin labels are read from column headings as suffices in the
    format e.g. "z_0.5_1.5".  If symmetric uncertainties are given for
    base-10 logarithmic luminosity function and conversion to linear values
    is needed, the larger asymmetric uncertainty on the linear luminosity
    function value is taken.

    """

    def __init__(self, measurement_file, uncertainty_file=None,
                 base10_log=True):

        self._lg_conversion = base10_log
        self._measurement_source_path = measurement_file
        self._uncertainty_source_path = uncertainty_file

        self._load_source_files()

        self._valid_data_points = ~np.isnan(self._measurements.flatten())
        if self._uncertainties is not None:
            self._valid_data_points &= ~np.isnan(self._uncertainties.flatten())
            if not self._lg_conversion:
                self._valid_data_points &= np.greater_equal(
                    np.abs(self._measurements.flatten()),
                    np.abs(self._uncertainties.flatten())
                )

    def __str__(self):

        return "LuminosityFunctionMeasurements({})".format(",".join(
            [
                "measurement_source={}".format(self._measurement_source_path),
                "uncertainty_source={}".format(self._uncertainty_source_path),
            ]
        ))

    def __getitem__(self, z_key):
        """Get luminosity function measurements and uncertainties (if
        available) for a specific redshift bin.

        Parameters
        ----------
        z_key: int, slice or str
            Slice or integer index or string representing redshift
            bins.  If a string, the accepted format is e.g. ``z=1.0``.

        Returns
        -------
        :class:`numpy.ndarray`, :class:`numpy.ndarray` or None
            Measurements and uncertainties for the redshift bin(s).

        """
        if isinstance(z_key, (int, slice)):
            z_idx = z_key
        else:
            try:
                z = float(str(z_key).replace(" ", "").lstrip("z="))
                z_idx = self.redshift_bins.index(z)
            except (TypeError, ValueError):
                raise KeyError(
                    "Non-existent redshift bin for '{}'.".format(z_key)
                )

        if self._uncertainties is not None:
            return self._measurements[z_idx], self._uncertainties[z_idx]
        return self._measurements[z_idx], None

    def get_statistics(self):
        """Return the empirical luminosity function data measurements
        and uncertainties as mean and variance statistics.

        Returns
        -------
        data_mean : :class:`numpy.ndarray`
            Luminosity function measurements as empirical mean.
        data_variance : :class:`numpy.ndarray` or None
            Luminosity function uncertainties as empirical variance,
            if available.

        Notes
        -----
        Data vectors are ordered by increasing redshift bins and
        within the same redshift bin ordered by increasing luminosity.

        """
        data_mean = self._measurements.flatten()[self._valid_data_points]

        if self._uncertainties is not None:
            data_variance = np.square(
                self._uncertainties.flatten()[self._valid_data_points]
            )
        else:
            data_variance = None

        return data_mean, data_variance

    def _load_source_files(self):

        # Process measurements.
        with open(self._measurement_source_path, 'r') as mfile:
            mheadings = process_header(mfile.readline(), skipcols=1)

        self.redshift_labels, self.redshift_bins = \
            self._extract_redshift_bins(mheadings)

        measurement_source = np.genfromtxt(
            self._measurement_source_path, unpack=True
        )

        self.luminosity_bins = measurement_source[0]

        measurement_array = measurement_source[1:]

        if len(measurement_array) != len(mheadings):
            raise DataSourceError(
                "Measurements do not match the number of headings ."
            )

        if not self._lg_conversion:
            for z_idx, col_name in enumerate(mheadings):
                if 'lg_' in col_name:
                    measurement_array[z_idx] = 10 ** measurement_array[z_idx]
                    mheadings[z_idx] = col_name.replace("lg_", "")

        self._measurements = measurement_array

        # Process uncertainties.
        if not self._uncertainty_source_path:
            self._uncertainties = None
        else:
            with open(self._uncertainty_source_path, 'r') as ufile:
                uheadings = process_header(ufile.readline(), skipcols=1)

            uncertainty_source = np.genfromtxt(
                self._uncertainty_source_path, unpack=True
            )

            luminosity_bin_matching_msg = (
                "Luminosity bins in measurement and uncertainty files "
                "do not match."
            )
            if len(self.luminosity_bins) != len(uncertainty_source[0]):
                raise DataSourceError(luminosity_bin_matching_msg)
            if not np.allclose(self.luminosity_bins, uncertainty_source[0]):
                warnings.warn(luminosity_bin_matching_msg, DataSourceWarning)

            uncertainty_array = uncertainty_source[1:].copy()

            if np.shape(uncertainty_array) != np.shape(measurement_array):
                raise DataSourceError(
                    "Uncertainty file data do not match measurement file data."
                )

            if not self._lg_conversion:
                # Optimistic wing of assymetric uncertainties.
                error_array = uncertainty_source[1:].copy()
                for z_idx, col_name in enumerate(uheadings):
                    if 'lg_' in col_name:
                        error_array[z_idx] = measurement_array[z_idx] \
                            * (1 - 10 ** (- error_array[z_idx]))
                        # Larger asymetric uncertainty is taken.
                        uncertainty_array[z_idx] = measurement_array[z_idx] \
                            * (10 ** uncertainty_array[z_idx] - 1)
                        uheadings[z_idx] = col_name.replace("lg_", "")
                self._errors = error_array

        self._uncertainties = uncertainty_array

    @staticmethod
    def _extract_redshift_bins(headings):

        bin_labels = sorted(set(map(
            r"${}$".format,
            [
                column.split("z_")[-1].replace("_", "<z<")
                for column in headings if 'z_' in column
            ]
        )))

        bin_centres = np.mean(
            [
                tuple(map(float, column.split("z_")[-1].split("_")))
                for column in headings if 'z_' in column
            ],
            axis=-1
        )

        return bin_labels, bin_centres


def _uniform_log_pdf(param_vals, param_ranges):

    if len(param_ranges) != len(param_vals):
        raise ValueError(
            "Number of parameter ranges does not match "
            "the number of parameters."
        )

    param_ranges = np.sort(np.atleast_2d(param_ranges), axis=-1)

    if any(np.less(param_vals, param_ranges[:, 0])) \
            or any(np.greater(param_vals, param_ranges[:, 1])):
        return - np.inf
    return 0.


def _normal_log_pdf(deviation_vector, covariance_matrix):

    deviation_vector = np.asarray(deviation_vector)

    if not all(np.isfinite(deviation_vector)):
        return - np.inf

    return - 1/2 * np.dot(
        deviation_vector, np.linalg.solve(covariance_matrix, deviation_vector)
    )


class LumFuncLikelihood(LumFuncMeasurements):
    """Luminosity function likelihood.

    Notes
    -----
    The built-in likelihood distribution is a multivariate normal
    approximation near the maximum point of the Poisson distribution that
    describes the tracer number count in redshift and luminosity bins,
    with different prescriptions of the luminosity function uncertainties
    to the diagonal covariance matrix of the multivariate normal
    distribution (see appendix B of [1]_).  The built-in prior
    distribution is multivariate uniform.

    .. [1] Pozzetti L. et al., 2016. A&A 590, A3.
       [arXiv: `1603.01453 <https://arxiv.org/abs/1603.01453>`_]


    Parameters
    ----------
    model_lumfunc : callable
        Luminosity function model.  Must return base-10 logarithmic values
        or accept `base10_log` boolean keyword argument.
    measurement_file : str or :class:`pathlib.Path`
        Luminosity function measurement file path.
    prior_file : str or :class:`pathlib.Path`
        Luminosity function model prior file path.  The prior parameter
        values (parameter ranges) may not matter, but the parameter names
        provided in the file do.
    model_constraint : callable or None, optional
        Additional model constraint(s) to be imposed on model parameters
        as a prior (default is `None`).
    uncertainty_file : str or :class:`pathlib.Path` *or None, optional*
        Luminosity function uncertainty file path (default is `None`).
        Ignored if `data_covariance` is provided.
    fixed_file : str or :class:`pathlib.Path`*or None, optional*
        Luminosity function model fixed parameter file path (default is
        `None`).  This covers any model parameter(s) not included in
        the prior.
    prescription : {'native', 'poisson', 'symlg', 'symlin'}, str, optional
        Gaussian likelihood approximation prescription (default is
        'poisson').
    model_options : dict or None, optional
        Additional parameters passed to the `model_lumfunc` for model
        evaluation (default is `None`).  This should not contain
        parametric luminosity function model parameters but only Python
        function implementation optional parameters (e.g.
        ``redshift_pivot=2.2`` for
        :func:`~horizonground.lumfunc_modeller.quasar_PLE_lumfunc`).
    covariance_matrix : float array_like or None, optional
        Covariance matrix for the multivariate normal likelihood
        approximation (default is `None`).  Its dimensions must match the
        length of the data vector flattened by redshift and
        luminosity bins.

    Attributes
    ----------
    data_points : list of float
        A vector of (luminosity, redshift) points for each valid
        luminosity function measurement.
    data_vector : float :class:`numpy.ndarray`
        Flattened luminosity function measurements for valid data points.
    prior, fixed : :class:`collections.OrderedDict` or None
        Ordered dictionary of varied prior parameter names and values or
        fixed parameter names and values.  The parameters are ordered
        as in the input files, so that the ordering of luminosity function
        arguments is consistent.

    """

    def __init__(self, model_lumfunc, measurement_file, prior_file,
                 uncertainty_file=None, fixed_file=None,
                 prescription='poisson', model_constraint=None,
                 model_options=None, covariance_matrix=None):

        # Set up likelihood treatment.
        self._prior_source_path = prior_file
        self._fixed_source_path = fixed_file
        self.prior, self.fixed = self._setup_prior()

        self._presciption = prescription
        if self._presciption in ['native']:
            self._lg_conversion = False
        elif self._presciption in ['poisson', 'symlg', 'symlin']:
            self._lg_conversion = True
        else:
            raise ValueError(
                f"Invalid Gaussian likelihood prescription: {prescription}."
            )

        # Set up model evaluation.
        self._model_lumfunc = model_lumfunc
        self._model_constraint = model_constraint
        self._model_options = model_options or {}

        if 'base10_log' in signature(self._model_lumfunc).parameters:
            self._model_options.update({'base10_log': self._lg_conversion})

        # Set up data processing.
        super().__init__(
            measurement_file, uncertainty_file=uncertainty_file,
            base10_log=self._lg_conversion
        )
        self.data_points = self._setup_data_points()
        self.data_vector, self._covariance = \
            self._get_moments(external_covariance=covariance_matrix)

    def __str__(self):

        return "LuminosityFunctionLikelihood({})".format(",".join(
            [
                "LF_model={}".format(self._model_lumfunc.__name__),
                "measurement_source={}".format(self._measurement_source_path),
                "uncertainty_source={}".format(self._uncertainty_source_path),
                "prior_source={}".format(self._prior_source_path),
                "fixed_source={}".format(self._fixed_source_path),
                "likelihood_approximant={}".format(self._presciption),
            ]
        ))

    def __call__(self, param_point, use_prior=False):
        """Evaluate the logarithmic likelihood at the model parameter
        point.

        Parameters
        ----------
        param_point : float array_like
            :attr:`model_lumfunc` parametric model parameters ordered in
            the same way as the prior parameters.
        use_prior : bool, optional
            If `True` (default is `False`), use the user-input prior range.

        Returns
        -------
        float
            Logarithmic likelihood value.

        """
        if isinstance(param_point, np.ndarray):
            param_point = param_point.tolist()
        else:
            param_point = list(param_point)

        # Check for prior.
        if use_prior:
            log_prior = _uniform_log_pdf(
                np.reshape(param_point, -1), list(self.prior.values())
            )
        else:
            log_prior = 0.

        if not np.isfinite(log_prior):
            return - np.inf

        # Check for model constraint.
        model_params = OrderedDict(zip(list(self.prior.keys()), param_point))
        if self.fixed is not None:
            model_params.update(self.fixed)

        try:
            within_constraint = self._model_constraint(model_params)
        except TypeError:
            within_constraint = True

        if not within_constraint:
            return - np.inf

        # Pass full parameter scope to the luminosity function model.
        model_params.update(self._model_options)

        model_vector = [
            self._model_lumfunc(*data_point, **model_params)
            for data_point in self.data_points
        ]

        # Form Gaussian approximant likelihood deviation vector.
        deviation_vector = np.subtract(model_vector, self.data_vector)
        covariance_matrix = self._covariance

        if self._presciption == 'poisson':
            deviation_vector = np.sqrt(
                2 * np.abs(np.exp(deviation_vector) - 1 - deviation_vector)
            )
        elif self._presciption == 'symlin':
            deviation_vector = np.exp(deviation_vector) - 1

        log_likelihood = _normal_log_pdf(deviation_vector, covariance_matrix)

        return log_prior + log_likelihood

    def _setup_data_points(self):

        # The primary axis is redshift, secondary luminosity.
        _data_points = list(map(
            lambda tup: tuple(reversed(tup)),
            iterproduct(self.redshift_bins, self.luminosity_bins)
        ))

        return list(iterfilter(_data_points, self._valid_data_points))

    def _setup_prior(self):

        with open(self._prior_source_path, 'r') as pfile:
            parameter_names = process_header(pfile.readline())

        prior_data = np.genfromtxt(self._prior_source_path, unpack=True)

        prior_ranges = list(map(tuple, prior_data))

        prior = OrderedDict(zip(parameter_names, prior_ranges))

        if self._fixed_source_path is None:
            fixed = None
        else:
            with open(self._fixed_source_path, 'r') as ffile:
                fixed_names = process_header(ffile.readline())
            fixed_values = np.genfromtxt(self._fixed_source_path, unpack=True)
            fixed = OrderedDict(zip(fixed_names, fixed_values))

        return prior, fixed

    def _get_moments(self, external_covariance=None):

        data_mean, data_var = self.get_statistics()

        if external_covariance is not None:
            data_covar = np.squeeze(external_covariance)
            if len(set(np.shape(data_covar))) > 1 \
                    or len(data_covar) != len(self.data_points):
                raise ValueError(
                    "`covariance_matrix` dimensions do not match data points: "
                    "({:d}, {:d}) versus {:d}."
                    .format(*np.shape(data_covar), len(self.data_points))
                )
            return data_mean, data_covar

        if data_var is None:
            raise ValueError(
                "Either `uncertainty_file` or `covariance_matrix` must be "
                "provided for setting the likelihood covariance matrix."
            )

        return data_mean, np.diag(data_var)



r"""
Clustering modification (:mod:`~horizonground.clustering_modification`)
===========================================================================

Compute modifications to the Newtonian power spectrum in the
distant-observer and plane-parallel limits.


Standard Kaiser RSD model
---------------------------------------------------------------------------

Redshift-space distortions induce anisotropy in the power spectrum
multipoles

.. math::

    \begin{align*}
        P_0(k, z) &=
            \left[
                b_1(z)^2 + \frac{2}{3} b_1(z) f(z) + \frac{1}{5} f(z)^2
            \right] P_\mathrm{m}(k, z) \,, \\
        P_2(k, z) &=
            \left[
                \frac{4}{3} b_1(z) f(z) + \frac{4}{7} f(z)^2
            \right] P_\mathrm{m}(k, z) \,, \\
        P_4(k, z) &= \frac{8}{35} f(z)^2 P_\mathrm{m}(k, z) \,
    \end{align*}

where :math:`P_\mathrm{m}` is the matter power spectrum, :math:`b_1(z)` is
the linear bias and :math:`f(z)` is the linear growth rate.

.. autosummary::

    standard_kaiser_factor


Non-Gaussianity scale-dependent modifications
---------------------------------------------------------------------------

Local primordial non-Gaussianty :math:`f_\mathrm{NL}` induces scale
dependence in the linear bias,

.. math::

    b_1(z) \mapsto b_1(z) + \Delta b_k(z) \,, \quad
    \Delta b(k, z) = f_\mathrm{NL} [b_1(z) - p] \frac{A(k, z)}{k^2} \,,

where the scale-dependence kernel is

.. math::

    A(k, z) = 3 \left( \frac{H_0}{\mathrm{c}} \right)^2
        \frac{1.27 \varOmega_\mathrm{m,0} \delta_\mathrm{c}}{D(z)T(k)} \,.

Here :math:`H_0` is the Hubble parameter at the current epoch
(in km/s/Mpc), :math:`\mathrm{c}` the speed of light,
:math:`\varOmega_\mathrm{m,0}` the matter density parameter at the current
epoch, and :math:`\delta_\mathrm{c}` the critical over-density in
spherical gravitational collapse.  The growth factor :math:`D(z)` is
normalised to unity at the current epoch (thus the numerical factor 1.27),
the transfer function :math:`T(k)` is normalised to unity as
:math:`k \to 0`, and :math:`p` is a tracer-dependent parameter.

Modifications to power spectrum multipoles as a result of local primordial
non-Gaussianty are

.. math::

    \begin{align*}
        \Delta P_0(k, z) &= \left[
            \left( 2 b_1 + \frac{2}{3} f \right) \Delta b(k, z)
            + \Delta b(k, z)^2
        \right] P_\mathrm{m}(k, z) \,, \\
        \Delta P_2(k, z) &=
            \frac{4}{3} f \Delta b(k, z) P_\mathrm{m}(k, z) \,.
    \end{align*}

.. autosummary::

    scale_dependence_kernel
    non_gaussianity_correction_factor


Relativistic corrections
---------------------------------------------------------------------------

Relativistic corrections to the Newtonian clustering mode

.. math::

    \delta(\mathbf{k}, z)
    = b_1(z) \delta_\mathrm{m}(\mathbf{k}, z)
        + g(z) v_{\parallel}(\mathbf{k}, z)
    = b_1(z) \delta_\mathrm{m}(\mathbf{k}, z)
        + \mathrm{i} \frac{\mathcal{H}}{k} g(z) f(z) \mu
        \delta_\mathrm{m}(\mathbf{k}, z)

are parametrised by the redshift-dependent, dimensionless quantity

.. math::

    g(z) = \frac{\mathcal{H}'}{\mathcal{H}^2}
        + 5s + \frac{2 - 5s}{\mathcal{H} \chi}
        - b_\mathrm{e}

with evolution bias :math:`b_\mathrm{e}(z)` and magnification bias
:math:`s(z)`, where :math:`v_{\parallel}` is the line-of-sight peculiar
velocity, :math:`\chi(z)` is the comoving distance and :math:`'` denotes
derivatives with respect to the conformal time. This can be written as the
sum of three contributions

.. math::

    g(z) = \underbrace{\left[
        1
        - \frac{3}{2} \frac{H_0^2}{H(z)^2} \varOmega_\mathrm{m,0} (1 + z)^3
        + \frac{2}{\mathcal{H}\chi}
    \right]}_{\text{background expansion}}
    \underbrace{- b_\mathrm{e}(z)}_{\text{evolution}}
    + \underbrace{
        5s(z) \left( 1 - \frac{1}{\mathcal{H}\chi} \right)
    }_{\text{magnification}} \,.

Modifications to power spectrum multipoles from the relativistic
corrections are

.. math::

    \begin{align*}
        \Delta P_0(k, z) &= \frac{1}{3} \frac{\mathcal{H}^2}{k^2}
            g(z)^2 f(z)^2 P_\mathrm{m}(k,z) \,,\\
        \Delta P_2(k, z) &= \frac{2}{3} \frac{\mathcal{H}^2}{k^2}
            g(z)^2 f(z)^2 P_\mathrm{m}(k,z) \,,
    \end{align*}

.. autosummary::

    relativistic_correction_func
    relativistic_correction_value
    relativistic_correction_factor

|

"""
# pylint: disable=no-name-in-module
import numpy as np
from astropy.constants import c
from nbodykit.lab import cosmology as nbk_cosmology

_SPEED_OF_LIGHT_IN_KM_PER_S = c.to('km/s').value
_SPHERICAL_COLLAPSE_CRITICAL_OVERDENSITY = 1.686
_GROWTH_FACTOR_NORMALISATION = 1.27

FIDUCIAL_COSMOLOGY = nbk_cosmology.Planck15
r""":class:`nbodykit.cosmology.Cosmology`: Default Planck15 cosmology.

"""


def standard_kaiser_factor(order, bias, redshift, cosmo=FIDUCIAL_COSMOLOGY):
    r"""Compute the standard Kaiser power spectrum multipoles as multiples
    of the matter power spectrum, i.e. :math:`P_\ell/P_\mathrm{m}`.

    Parameters
    ----------
    order : int
        Order of the multipole, ``order >= 0``.
    bias : float
        Scale-independent linear bias at `redshift`.
    redshift : float
        Redshift.
    cosmo : :class:`nbodykit.cosmology.Cosmology`, optional
        Cosmological model (default is ``FIDUCIAL_COSMOLOGY``).

    Returns
    -------
    factor : float
        Power spectrum multipoles as multiples of the matter power
        spectrum.

    """
    b_1 = bias

    f = cosmo.scale_independent_growth_rate(redshift)

    if order == 0:
        factor = b_1 ** 2 + 2./3. * f * b_1 + 1./5. * f**2
    elif order == 2:
        factor = 4./3. * f * b_1 + 4./7. * f**2
    elif order == 4:
        factor = 8./35. * f**2
    else:
        factor = 0.

    return factor


def scale_dependence_kernel(redshift, cosmo=FIDUCIAL_COSMOLOGY):
    r"""Return the scale-dependence kernel :math:`A(k, z)` in the presence
    of local primordial non-Gaussianity.

    Parameters
    ----------
    redshift : float
        Redshift.
    cosmo : :class:`nbodykit.cosmology.Cosmology`, optional
        Cosmological model (default is ``FIDUCIAL_COSMOLOGY``).

    Returns
    -------
    callable
        Scale-dependence kernel as a function of wavenumber (in
        :math:`h/\mathrm{Mpc}`).

    """
    numerical_constants = 3 * _SPHERICAL_COLLAPSE_CRITICAL_OVERDENSITY \
        * cosmo.Om0 * _GROWTH_FACTOR_NORMALISATION \
        * (FIDUCIAL_COSMOLOGY.H0 / _SPEED_OF_LIGHT_IN_KM_PER_S)**2

    transfer_function = nbk_cosmology.power.transfers.CLASS(cosmo, redshift)

    return lambda k: numerical_constants / transfer_function(k)


def non_gaussianity_correction_factor(wavenumber, order, local_png, bias,
                                      redshift, cosmo=FIDUCIAL_COSMOLOGY,
                                      tracer_p=1.):
    r"""Compute modifications to the power spectrum multipoles by local
    primordial non-Gaussianity as multiples of the matter power spectrum,
    i.e. :math:`\Delta P_\ell/P_\mathrm{m}`.

    Parameters
    ----------
    wavenumber : float, array_like
        Wavenumber (in :math:`h/\mathrm{Mpc}`).
    order : int
        Order of the multipole, ``order >= 0``.
    local_png : float
        Local primordial non-Gaussianity.
    bias : float
        Scale-independent linear bias at `redshift`.
    redshift : float
        Redshift.
    cosmo : :class:`nbodykit.cosmology.Cosmology`, optional
        Base cosmological model (default is ``FIDUCIAL_COSMOLOGY``).
    tracer_p : float, optional
        Tracer parameter (default is 1.).

    Returns
    -------
    factor : float :class:`numpy.ndarray`
        Power spectrum multipole modifications as multiples of the matter
        power spectrum.

    """
    f_nl, p = local_png, tracer_p
    b_1 = bias(redshift) if callable(bias) else bias

    f = cosmo.scale_independent_growth_rate(redshift)

    delta_b = f_nl * (b_1 - p) \
        * scale_dependence_kernel(redshift, cosmo=cosmo)(wavenumber) \
        / wavenumber**2

    if order == 0:
        factor = (2 * b_1 + 2./3. * f) * delta_b + delta_b ** 2
    elif order == 2:
        factor = 4./3. * f * delta_b
    else:
        factor = 0.

    return factor


def relativistic_correction_func(cosmo=FIDUCIAL_COSMOLOGY, geometric=True,
                                 evolution_bias=None, magnification_bias=None):
    r"""Return the relativistic correction function as
    :math:`\mathcal{H} g(z)`.

    Parameters
    ----------
    cosmo : :class:`nbodykit.cosmology.Cosmology`, optional
        Cosmological model (default is ``FIDUCIAL_COSMOLOGY``).
    geometric : bool, optional
        If `True` (default), include geometric contributions from the
        background expansion.
    evolution_bias, magnification_bias : callable or None, optional
        Evolution bias or magnification bias as a function of redshift
        (default is `None`).

    Returns
    -------
    callable
        Relativistic correction function as a function of redshift
        (in :math:`h`/Mpc).

    """
    astropy_cosmo = cosmo.to_astropy()

    aH = lambda z: astropy_cosmo.scale_factor(z) \
        * astropy_cosmo.H(z).value / _SPEED_OF_LIGHT_IN_KM_PER_S
    chi = lambda z: astropy_cosmo.comoving_distance(z).value

    if geometric:
        geometric_term = lambda z: \
            1 - 3./2. * astropy_cosmo.Om0 \
            * (astropy_cosmo.H0 / astropy_cosmo.H(z)) ** 2 \
            * (1 + z) ** 3
    else:
        geometric_term = lambda z: 0.

    if evolution_bias is None:
        evolution_term = lambda z: 0.
    else:
        evolution_term = lambda z: 2 / (aH(z) * chi(z)) - evolution_bias(z)

    if magnification_bias is None:
        lensing_term = lambda z: 0.
    else:
        lensing_term = lambda z: \
            5 * magnification_bias(z) * (1 - 1 / (aH(z) * chi(z)))

    return np.vectorize(
        lambda z: aH(z) / astropy_cosmo.h
        * (geometric_term(z) + evolution_term(z) + lensing_term(z))
    )


def relativistic_correction_value(redshift, cosmo=FIDUCIAL_COSMOLOGY,
                                  geometric=True, evolution_bias=None,
                                  magnification_bias=None):
    r"""Evaluate the relativistic correction function
    :math:`\mathcal{H} g(z)` at a given redshift.

    Parameters
    ----------
    redshift : float
        Redshift.
    cosmo : :class:`nbodykit.cosmology.Cosmology`, optional
        Base cosmological model (default is ``FIDUCIAL_COSMOLOGY``).
    geometric : bool, optional
        If `True` (default), include geometric contributions from the
        background expansion.
    evolution_bias, magnification_bias : float or None, optional
        Evolution bias or magnification bias evaluated at input `redshift`
        (default is `None`).

    Returns
    -------
    correction_value : float
        Relativistic correction function value (in :math:`h`/Mpc)
        at `redshift`.

    """
    astropy_cosmo = cosmo.to_astropy()

    aH = astropy_cosmo.scale_factor(redshift) \
        * astropy_cosmo.H(redshift).value / _SPEED_OF_LIGHT_IN_KM_PER_S
    chi = astropy_cosmo.comoving_distance(redshift).value

    correction_value = 0.
    if geometric:
        correction_value += 1 - 3./2. * astropy_cosmo.Om0 \
            * (astropy_cosmo.H0 / astropy_cosmo.H(redshift)) ** 2 \
            * (1 + redshift) ** 3
    if evolution_bias:
        correction_value += 2 / (aH * chi) - evolution_bias
    if magnification_bias:
        correction_value += 5 * magnification_bias * (1 - 1 / (aH * chi))

    return aH / astropy_cosmo.h  * correction_value


def relativistic_correction_factor(wavenumber, order, redshift,
                                   correction_value=None,
                                   cosmo=FIDUCIAL_COSMOLOGY,
                                   geometric=True,
                                   evolution_bias=None,
                                   magnification_bias=None):
    r"""Compute modifications to the power spectrum multipoles by
    relativistic corrections as multiples of the matter power spectrum,
    i.e. :math:`\Delta P_\ell/P_\mathrm{m}`.

    Parameters
    ----------
    wavenumber : float, array_like
        Wavenumber (in :math:`h/\mathrm{Mpc}`).
    order : int
        Order of the multipole, ``order >= 0``.
    redshift : float
        Redshift.
    correction_value : float or None, optional
        If not `None` (default), this is directly used as
        :math:`\mathcal{H} g(z)` at `redshift` in calculations, and
        `cosmo`, `geometric`, `evolution_bias` and `magnification_bias`
        are ignored.
    cosmo : :class:`nbodykit.cosmology.Cosmology`, optional
        Cosmological model (default is ``FIDUCIAL_COSMOLOGY``).
    geometric : bool, optional
        If `True` (default), include geometric contributions from the
        background expansion.
    evolution_bias : float or callable or None, optional
        Evolution bias as a function of redshift, or evaluated at input
        `redshift` (default is `None`).  If callable/float,
        `magnification_bias` must also be callable/float.
    magnification_bias : float or callable or None, optional
        Magnification bias as a function of redshift, or evaluated at input
        `redshift` (default is `None`).  If callable/float,
        `evolution_bias` must also be callable/float.

    Returns
    -------
    factor : float :class:`numpy.ndarray`
        Power spectrum multipoles as multiples of the matter power
        spectrum.

    """
    if correction_value is None:
        if callable(evolution_bias) and callable(magnification_bias):
            correction_function = relativistic_correction_func(
                cosmo=cosmo,
                geometric=geometric,
                evolution_bias=evolution_bias,
                magnification_bias=magnification_bias
            )
            correction_value = correction_function(redshift)
        elif isinstance(evolution_bias, float) \
            and isinstance(magnification_bias, float):
            correction_value = relativistic_correction_value(
                redshift,
                cosmo=cosmo,
                geometric=geometric,
                evolution_bias=evolution_bias,
                magnification_bias=magnification_bias
            )
        elif evolution_bias is None and magnification_bias is None:
            correction_value = relativistic_correction_value(
                redshift, cosmo=cosmo, geometric=geometric
            )
        else:
            raise TypeError(
                "`evolution_bias` and `magnification_bias` must be "
                "both callable or float or None."
            )

    modification = correction_value ** 2 \
        * cosmo.scale_independent_growth_rate(redshift) ** 2 \
        / wavenumber ** 2

    if order == 0:
        factor = 1./3. * modification
    elif order == 2:
        factor = 2./3. * modification
    else:
        factor = 0.

    return factor



# pylint: disable=missing-module-docstring
import setuptools

with open("README.md", 'r') as readme:
    long_description = readme.read()

with open("requirements.txt", 'r') as dependencies:
    requirements = [pkg.strip() for pkg in dependencies]

VERSION = "0.1.0"

setuptools.setup(
    name="HorizonGRound",
    version=VERSION,
    author="Mike S Wang",
    author_email="mike.wang@port.ac.uk",
    license="GPLv3",
    description=(
        "Forward-modelling of relativistic effects "
        "from tracer luminosity function.",
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MikeSWang/HorizonGRound/tree/v{}".format(VERSION),
    packages=['horizonground', 'horizonground.tests'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires='>=3.6',
    project_urls={
        "Documentation": "https://mikeswang.github.io/HorizonGRound/",
        "Source": "https://github.com/MikeSWang/HorizonGRound/tree/v{}"\
            .format(VERSION),
    },
)
